import random
import os
import ast
import numpy
import sys
from functools import reduce
from . import SimUtils
from . import TileType
from . import Frontier
from . import TileType
from . import SimStep

GRID_WIDTH  = 100
GRID_HEIGHT = 100
GROWTH_INC  = 200

EMPTY_LOC           = 0
FRONTIER_LOC        = 1
FRONTIER_LOC_OFF    = 2

TILESET_LOAD_SUCCESS = 0
TILESET_LOAD_FAIL = 1
TILESET_LOAD_FAIL_NEEDS_3D = 2
SCRIPT_LOAD_FAIL = 3

MAX_COORD = 9999999

#===============================================================================
# N  = 0
# NE = 1
# E  = 2
# SE = 3
# S  = 4
# SW = 5
# W  = 6
# NW = 7
#===============================================================================





class Assembly:
    def __init__(self, width=GRID_WIDTH, height=GRID_HEIGHT, tileWidths=100, tileHeights=100, minX=None, maxX=None, minY=None, maxY=None):
        #globalApp.OutputMessage("Creating assembly with width=" + str(width) + " and height=" + str(height))
        self.ALLDIRS = 8
        self.DIRSTRS = ["N", "NE", "E", "SE", "S", "SW", "W", "NW"]
        self.NBRS = [(0,1),(1,1),(1,0),(1,-1),(0,-1),(-1,-1),(-1,0),(-1,1)]
        self.OPPDIRS = [4, 5, 6, 7, 0, 1, 2, 3]

        if (minX is None) or (maxX is None) or (minY is None) or (maxY is None):
            self.width = width
            self.height = height
            self.minX = 1 - int(width/2)
            self.maxX = int(width/2)
            self.minY =  1 - int(height/2)
            self.maxY = int(height/2)
        else:
            self.width = maxX - minX + 1
            self.height = maxY - minY + 1
            self.minX = minX
            self.maxX = maxX
            self.minY = minY
            self.maxY = maxY
        self.growthIncrement = GROWTH_INC
        #self.grid = [[None for x in range(width+1)] for y in range(height+1)]
        #self.grid = [[None] * self.height for i in range(self.width)]
        self.grid = numpy.full(shape=(self.height, self.width), dtype=object, fill_value=None)
        self.tileTypes = dict()  
        self.tileWidths = tileWidths
        self.tileHeights = tileHeights      
        self.seedPatch = []
        self.temperature = 2
        self.frontier = set()
        self.disabledFrontier = set()
        self.numTiles = 0
        self.tilesetDefFile = None
        self.systemDefFile = None
        self.tileSystemDef = None
        self.bKeepCache = True
        self.cache = []
        self.currCacheLoc = -1
        self.tileSideHashtables = []
        self.ResetMaxCoords()
        #globalApp.OutputMessage("Created new grid with (minX,minY) = " + str((self.minX,self.minY)) + " and (maxX,maxY) = " + str((self.maxX,self.maxY)))

    def ResetMaxCoords(self):
        self.minTiledX =  MAX_COORD
        self.maxTiledX = -MAX_COORD
        self.minTiledY =  MAX_COORD
        self.maxTiledY = -MAX_COORD
        self.minTiledZ =  MAX_COORD
        self.maxTiledZ = -MAX_COORD

    def GetWidth(self):
        return self.width

    def GetHeight(self):
        return self.height

    def GetNumTiles(self):
        return self.numTiles

    def GrowSpace(self, diff=(0,0)):
        newWidth = self.width
        newHeight = self.height
        newMinX = self.minX
        newMaxX = self.maxX
        newMinY = self.minY
        newMaxY = self.maxY
        if (diff[0] < 0):
            growthAmount = abs(diff[0]) + self.width
            if growthAmount < self.growthIncrement:
                growthAmount = self.growthIncrement
            newWidth = self.width + growthAmount
            newMinX = self.minX - growthAmount#(diff[0] + self.width)
            newLocs = numpy.full(shape=(self.height, growthAmount), dtype=object, fill_value=None)
            self.grid = numpy.concatenate((newLocs, self.grid), axis=1)
            #print("self.width: " + str(self.width) + ", self.minX: " + str(self.minX) + ", newWidth: " + str(newWidth) + ", newMinX: " + str(newMinX))
        elif (diff[0] > 0):
            growthAmount = abs(diff[0]) + self.width
            if growthAmount < self.growthIncrement:
                growthAmount = self.growthIncrement
            newWidth = self.width + growthAmount
            newMaxX = self.maxX + growthAmount#(diff[0] + self.width)
            newLocs = numpy.full(shape=(self.height, growthAmount), dtype=object, fill_value=None)
            self.grid = numpy.concatenate((self.grid, newLocs), axis=1)
            #print("self.width: " + str(self.width) + ", self.maxX: " + str(self.maxX) + ", newWidth: " + str(newWidth) + ", newMaxX: " + str(newMaxX))
        if (diff[1] < 0):
            growthAmount = abs(diff[1]) + self.height
            if growthAmount < self.growthIncrement:
                growthAmount = self.growthIncrement
            newHeight = self.height + growthAmount
            newMinY = self.minY - growthAmount#(diff[1] + self.height)
            newLocs = numpy.full(shape=(growthAmount, self.width), dtype=object, fill_value=None)
            self.grid = numpy.concatenate((newLocs, self.grid), axis=0)
            #print("self.height: " + str(self.height) + ", self.minY: " + str(self.minY) + ", newHeight: " + str(newHeight) + ", newMinY: " + str(newMinY))
        elif (diff[1] > 0):
            growthAmount = abs(diff[1]) + self.height
            if growthAmount < self.growthIncrement:
                growthAmount = self.growthIncrement
            newHeight = self.height + growthAmount
            newMaxY = self.maxY + growthAmount#(diff[1] + self.height)
            newLocs = numpy.full(shape=(growthAmount, self.width), dtype=object, fill_value=None)
            self.grid = numpy.concatenate((self.grid, newLocs), axis=0)
            #print("self.height: " + str(self.height) + ", self.maxY: " + str(self.maxY) + ", newHeight: " + str(newHeight) + ", newMaxY: " + str(newMaxY))
        #SimUtils.globalApp.OutputMessage("Growing assembly from (width,height) " + str((self.width,self.height)) + " to " + str((newWidth,newHeight)))
        
        self.width = newWidth
        self.height = newHeight
        self.minX = newMinX
        self.maxX = newMaxX
        self.minY = newMinY
        self.maxY = newMaxY

#===============================================================================
#         newAssembly = Assembly(minX=newMinX, maxX=newMaxX, minY=newMinY, maxY=newMaxY)
# 
#         for i in range(self.minX, self.maxX+1):
#             for j in range(self.minY, self.maxY+1):
#                 loc = self.GetLoc((i,j))
#                 if loc is not None:
#                     newAssembly.SetLoc((i,j), loc.type)
#         del self.grid[:]
#         self.grid = newAssembly.grid
#         self.width = newAssembly.width
#         self.height = newAssembly.height
#         self.minX = newAssembly.minX
#         self.maxX = newAssembly.maxX
#         self.minY = newAssembly.minY
#         self.maxY = newAssembly.maxY
#===============================================================================
        
    def ClearSpace(self):
        newWidth = GRID_WIDTH
        newHeight = GRID_HEIGHT
        newAssembly = Assembly(newWidth, newHeight)
        del self.grid
        self.grid = newAssembly.grid
        self.width = newAssembly.width
        self.height = newAssembly.height
        self.minX = newAssembly.minX
        self.maxX = newAssembly.maxX
        self.minY = newAssembly.minY
        self.maxY = newAssembly.maxY
        self.frontier = set()
        self.disabledFrontier = set()
        self.numTiles = 0
        self.ResetMaxCoords()
        
    def CheckBounds(self, coords):
        xDiff = 0
        yDiff = 0
        if (coords[0] < self.minX):
            xDiff = coords[0] - self.minX
        elif (coords[0] > self.maxX):
            xDiff = coords[0] - self.maxX
        if (coords[1] < self.minY):
            yDiff = coords[1] - self.minY
        elif (coords[1] > self.maxY):
            yDiff = coords[1] - self.maxY
        return (xDiff,yDiff)

    def GetLoc(self, coords, growIfNeeded=False, includeFrontier=False):
        diff = self.CheckBounds(coords)
        if (diff[0] != 0) or (diff[1] != 0):
            if growIfNeeded == True:
                self.GrowSpace(diff)
            else:
                return None
        try:
            #loc = self.grid[coords[0]-self.minX][coords[1]-self.minY]
            loc = self.grid[coords[1]-self.minY][coords[0]-self.minX]
        except:
            print("Exception getting coords " + str(coords) + " with (minX,minY) = " + str((self.minX,self.minY)) + ", (maxX,maxY) = " + str((self.maxX,self.maxY)) + ", and (width,height) = " + str((self.width,self.height)))
            raise
        if includeFrontier == False and (loc == FRONTIER_LOC or loc == FRONTIER_LOC_OFF):
            return None
        else:
            return loc
    
    def UpdateMaxCoords(self, coords):
        if coords[0] < self.minTiledX:
            self.minTiledX = coords[0]
        if coords[0] > self.maxTiledX:
            self.maxTiledX = coords[0]
        if coords[1] < self.minTiledY:
            self.minTiledY = coords[1]
        if coords[1] > self.maxTiledY:
            self.maxTiledY = coords[1]
        if len(coords) > 2:
            if coords[2] < self.minTiledZ:
                self.minTiledZ = coords[2]
            if coords[2] > self.maxTiledZ:
                self.maxTiledZ = coords[2]
        
    def SetLoc(self, coords, ttype, growIfNeeded=False):
        diff = self.CheckBounds(coords)
        if (diff[0] != 0) or (diff[1] != 0):
            if growIfNeeded == True:
                self.GrowSpace(diff)
            else:
                return None
        if ttype == FRONTIER_LOC or ttype == FRONTIER_LOC_OFF:
            #self.grid[coords[0]-self.minX][coords[1]-self.minY] = ttype
            self.grid[coords[1]-self.minY][coords[0]-self.minX] = ttype
        else:
            #self.grid[coords[0]-self.minX][coords[1]-self.minY] = TileType.Tile(ttype, coords)
            self.grid[coords[1]-self.minY][coords[0]-self.minX] = TileType.Tile(ttype, coords)
            self.UpdateMaxCoords(coords)
        return True
        
    def ClearLoc(self, coords):
        if (coords[0] < self.minX) or (coords[0] > self.maxX):
            return False
        if (coords[1] < self.minY) or (coords[1] > self.maxY):
            return False
        #self.grid[coords[0]-self.minX][coords[1]-self.minY] = None
        self.grid[coords[1]-self.minY][coords[0]-self.minX] = None
        return True
        
    def AddTile(self, ttype, coords):
        loc = self.GetLoc(coords)
        if not loc is None:
            SimUtils.PrintDebug("ERROR: Trying to add another tile at occupied location " + str(coords), SimUtils.ERROR_MSG)
            return (False, None, True)
        if self.SetLoc(coords, ttype, growIfNeeded=True) == False:
            return (False, None, True)
        floc = Frontier.FrontierLocation(coords,"")
        if floc in self.frontier:
            self.frontier.remove(floc)
        emptyNbrLocs = self.GetNbrLocs(coords, True)
        newFrontierLocs = []
        #print("unoccupied nbrs of " + str(coords) + ": " + str(emptyNbrLocs))
        canKeepGoing = True
        for n in emptyNbrLocs:   # loop over unoccupied nbr locations to find occupied nbrs of each and build a list of fitting types
                fLoc = self.AddToFrontierIfNecessary(n)
                if not fLoc is None:
                    newFrontierLocs.append(fLoc)
                    if len(fLoc.fittingTypeNames) > 1:
                        if (SimUtils.globalApp.nondetMode == SimUtils.SIMFW_NONDET_REPORT) or (SimUtils.globalApp.nondetMode == SimUtils.SIMFW_NONDET_STOP):
                            SimUtils.globalApp.OutputMessage("Nondeterminism detected!  The following tile types can be placed in location " + str(n) + ": " + str(fLoc.fittingTypeNames))
                            if SimUtils.globalApp.nondetMode == SimUtils.SIMFW_NONDET_STOP:
                                canKeepGoing = False
        self.numTiles = self.numTiles + 1
        return (True, newFrontierLocs, canKeepGoing)

    def AddToFrontierIfNecessary(self, coords, growIfNeeded=True):
        fittingNbrSet = self.GetFittingTilesForLoc(coords, growIfNeeded)
        if fittingNbrSet is None:
            return None
        else:
            fLoc = Frontier.FrontierLocation(coords, fittingNbrSet)
            self.frontier.discard(fLoc)
            self.frontier.add(fLoc)
            retVal = None
            if fLoc in self.disabledFrontier:
                self.ClearLoc(coords)
            else:
                self.SetLoc(coords, FRONTIER_LOC, growIfNeeded)
                retVal = fLoc
            return retVal

    def RemoveTile(self, coords):
        if not self.GetLoc(coords) is None:
            self.ClearLoc(coords)
            self.numTiles = self.numTiles - 1

            # Check to see if this location should now be added to the frontier
            self.AddToFrontierIfNecessary(coords)

            # Check to see if nbrs were in the frontier but should now be removed
            nbrLocs = self.GetNbrLocs(coords, True)
            for n in nbrLocs:
                fLoc = Frontier.FrontierLocation(n, None)
                self.frontier.discard(fLoc)
                self.ClearLoc(n)
                self.AddToFrontierIfNecessary(n)
            return True
        else:
            print("ERROR: Trying to remove nonexistent node at location" + str(coords))
            return False


    def GetNbrLocs(self, coords, bEmpty=False):
        retlist = []
        for n in self.NBRS:
            nbrCoords = (coords[0]+n[0], coords[1]+n[1])
            if bEmpty == False:
                retlist.append(nbrCoords)
            elif self.GetLoc(nbrCoords) is None:
                retlist.append(nbrCoords)
        return retlist


    def GetOccupiedNbrLocs(self, coords):
        retlist = []
        for n in self.NBRS:
            nbrCoords = (coords[0]+n[0], coords[1]+n[1])
            nbrNode = self.GetLoc(nbrCoords)
            if not nbrNode is None:
                retlist.append(nbrNode)
        return retlist
    

    # This is just a stubbed out function to show how to gather up info about neighbors, and must be completed
    # to work correctly for the system being simulated
    def GetFittingTilesForLoc(self, coords, growIfNeeded=True):
        retSet = None
        count = 0
        for n in self.NBRS:
            nbrCoords = (coords[0]+n[0], coords[1]+n[1])
            nbrTile = self.GetLoc(nbrCoords, growIfNeeded)
            if not nbrTile is None:
                oppDir = (count + 4) % 8
                #TODO: Fill in code based on the system being simulated to determine what tile types can go here!!!
            count = count + 1
        return retSet

    def StepForward(self):
        if self.currCacheLoc < len(self.cache) - 1:
            self.currCacheLoc = self.currCacheLoc + 1
            cacheLoc = self.cache[self.currCacheLoc]
            for at in cacheLoc.addedTiles:
                if self.SetLoc(cacheLoc.addedTiles[0].coords, cacheLoc.addedTiles[0].type, growIfNeeded=True) == True:
                    self.numTiles = self.numTiles + 1
                else:
                    print("ERROR in StepForward trying to use cached step")
            for rt in cacheLoc.removedTiles:
                self.RemoveTile(rt.coords)
            for afl in cacheLoc.addedFrontierLocs:
                self.AddToFrontierIfNecessary(afl.coords)
            for rfl in cacheLoc.removedFrontierLocs:
                self.frontier.remove(rfl)
            return ((cacheLoc.addedTiles, cacheLoc.removedTiles, cacheLoc.addedFrontierLocs, cacheLoc.removedFrontierLocs), True)
        # Pick a spot from the frontier, add a tile
        elif len(self.frontier) == 0:
            SimUtils.globalApp.OutputMessage("The frontier is empty!")
            return None
        else:
            enabledFrontier = self.frontier
            if len(self.disabledFrontier) > 0:
                enabledFrontier = self.frontier - self.disabledFrontier
                if len(enabledFrontier) == 0:
                    SimUtils.globalApp.OutputMessage("There are no enabled frontier locations!")
                    return None
            frontierLoc = random.sample(enabledFrontier, 1)[0]
            typeName = frontierLoc.GetRandomFittingTileTypeName()
            added = self.AddTile(self.tileTypes[typeName], frontierLoc.coords)
            if added[0] == True:
                newTile = TileType.Tile(self.tileTypes[typeName], frontierLoc.coords)
                self.AddCacheEntry(addedTiles=[newTile], addedFrontierLocs=added[1], removedFrontierLocs=[frontierLoc])
                SimUtils.PrintDebug("Added a tile of type " + typeName + " to location " + str(frontierLoc.coords), SimUtils.MSG_MSG)
                return (([newTile], [], added[1], []), added[2])
            SimUtils.globalApp.OutputMessage("Error adding tile to location" + str(frontierLoc.coords))
            return None
                
    def AddCacheEntry(self, addedTiles=[], removedTiles=[], addedFrontierLocs=[], removedFrontierLocs=[]):
        if self.bKeepCache == True:
            if (self.currCacheLoc > -1) and (self.currCacheLoc < (len(self.cache))-1):
                self.cache = self.cache[0:self.currCacheLoc+1]
            self.cache.append(SimStep.SimStep(addedTiles, removedTiles, addedFrontierLocs, removedFrontierLocs))
            self.currCacheLoc = len(self.cache) - 1

    def RewindCache(self):
        if self.bKeepCache == True:
            if self.currCacheLoc == -1:
                return (False, None)
            if self.currCacheLoc >= 0:
                cacheLoc = self.cache[self.currCacheLoc]
                for at in cacheLoc.addedTiles:
                    self.RemoveTile(at.coords)
                for rt in cacheLoc.removedTiles:
                    self.AddTile(rt.type, rt.coords)
                for af in cacheLoc.addedFrontierLocs:
                    self.frontier.discard(af)
                    loc = self.GetLoc(at.coords, includeFrontier=True)
                    if (loc == FRONTIER_LOC) or (loc == FRONTIER_LOC_OFF):
                        self.ClearLoc(af.coords)
                for rf in cacheLoc.removedFrontierLocs:
                    self.frontier.add(rf)
                    self.SetLoc(at.coords, FRONTIER_LOC)
                self.currCacheLoc = self.currCacheLoc - 1
                
                return (True, (cacheLoc.removedTiles, cacheLoc.addedTiles, cacheLoc.removedFrontierLocs, cacheLoc.addedFrontierLocs))

    def PlaceSeed(self):
        SimUtils.globalApp.OutputMessage("Placing seed patch: " + str(self.seedPatch))
        minX = minY =  999999999
        maxX = maxY = -999999999
        for seedTile in self.seedPatch:
            self.AddTile(self.tileTypes[seedTile[0]], seedTile[1])
            x = int(seedTile[1][0])
            y = int(seedTile[1][1])
            if x < minX:
                minX = x
            if x > maxX:
                maxX = x
            if y < minY:
                minY = y
            if y > maxY:
                maxY = y
        aveX = minX + int((maxX - minX)/2)
        aveY = minY + int((maxY - minY)/2)
        return (aveX, aveY)
        
    # This is just a stubbed out version of a function to load an assembly system from file(s).  It must be completed
    # for your particular systems and model.
    def LoadTileSystem(self, fileName):
        if len(fileName) < 1:
            return False
        elif os.path.isabs(fileName) == False:
            fileName = "./" + fileName
            fileName = os.path.abspath(fileName)

        #try:
        os.stat(fileName)
        rFile = open(fileName, "r")
        SimUtils.globalApp.OutputMessage("Using tile system definition file '" + fileName + "'")
        lines = rFile.readlines()
        rFile.close()
        
        self.systemDefFile = fileName
        self.tileSystemDef = ast.literal_eval(lines[0])
        print("Read tile system " + str(self.tileSystemDef))

        self.tileTypes = dict()
        self.tileSideHashtables = []
        self.seedPatch = []
        self.frontier = set()
        self.enabledFrontier = set()
               
        SimUtils.globalApp.OutputMessage("Loaded tile set with size: " + str(len(self.tileTypes)))
        mid = self.PlaceSeed()
        return mid
    
    # Create a hash table for each tile side mapping glue labels to tile types for
    # quicker identification of fitting tile types for frontier locations
    def MakeTileSideHashtables(self):
        if len(self.tileTypes) < 1:
            return

        someTT = self.tileTypes[random.choice(list(self.tileTypes.keys()))]
        numSides = len(someTT.sides)
        indices = [x for x in range(numSides)]
        if type(someTT.sides) is dict:
            indices = list(someTT.sides.keys())
        #print("Num sides: " + str(numSides) + ", OPPDIRS: " + str(self.OPPDIRS) + ", indices: " + str(indices))
            
        self.tileSideHashtables = [None for n in range(numSides)]
        
        for i in range(numSides):
            hashTable = dict()
            for ttName in self.tileTypes:
                tt = self.tileTypes[ttName]

                if i < len(tt.sides):
                    if (len(tt.sides) > indices[self.OPPDIRS[i]]) and (not tt.sides[indices[self.OPPDIRS[i]]].strength == 0):
                        #glueHash = tt.sides[indices[i]].label + str(tt.sides[indices[i]].strength)
                        glueHash = tt.sides[indices[self.OPPDIRS[i]]].Hash()
                        if not glueHash in hashTable:
                            hashTable[glueHash] = set()
                        hashTable[glueHash].add(ttName)
            self.tileSideHashtables[self.OPPDIRS[i]] = hashTable
        #print("hash tables: " + str(self.tileSideHashtables))

    def SetCurrentAssemblyAsSeed(self):
        self.seedPatch = []
        for i in range(self.minX, self.maxX+1):
            for j in range(self.minY, self.maxY+1):
                loc = self.GetLoc((i,j))
                if loc is not None:
                    self.seedPatch.append((loc.type.name, loc.coords))
        self.ClearSpace()
        self.frontier = set()
        self.PlaceSeed()
        
    def AssemblyToString(self, tileList=None):
        retStr = ""
        for tile in tileList:
            retStr = retStr + tile.type.name + " " + str(tile.coords[0]) + " " + str(tile.coords[1])
            if len(tile.coords) > 2:
                retStr = retStr + " " + str(tile.coords[2])
            retStr = retStr + "\n"
        return retStr
        
    def AssemblyFromString(self, assemblyString):
        print("AssemblyFromString needs to be implemented!!!")
        return []
        
    def SaveAssembly(self, fileName):
        wFile = open(fileName, "w")
        assemblyString = self.AssemblyToString()
        wFile.write(assemblyString)
        wFile.close()
        SimUtils.globalApp.OutputMessage("Saved assembly to file '" + fileName + "'")
        
    def SaveTileset(self, fileName, tileSet=None):
        outFile = open(fileName, "w")
        if outFile is None:
            return False
        tilesToSave = self.tileTypes
        if not tileSet is None:
            tilesToSave = tileSet
        if type(tilesToSave) is dict:
            for ttName in tilesToSave:
                ttype = tilesToSave[ttName]
                outFile.write(ttype.Output() + "\n")
        else:
            for ttype in tilesToSave:
                outFile.write(ttype.Output() + "\n")
            
        outFile.close()
        return True

    def GetAllSubsets(self, lst):
        return reduce(lambda result, x: result + [subset + [x] for subset in result],
                      lst, [[]])
